@@ -24,11 +24,29 @@ Cond <- function(A) {
   d[1] / d[length(d)]
 }
 
-high.rank.rate <- function(F, L = (N + 1) %/% 2, rank, ...) {
-  N <- length(F); K <- N - L + 1
-  ss <- ssa(F, L, neig = min(rank + 1, L, K), ...)
+# TODO move this functionality to `clone.ssa()`
+.remove.ossa.pssa <- function(class) {
+  class <- class[(class != "ossa") & !grepl("^pssa", class)]
+}
 
-  1 - sum(.sigma(ss)[seq_len(rank)]^2) / wnorm(F, L)^2
+.clone.with.new.series <- function(x, F, ...) {
+  # Get conversion
+  conversion <- .inner.fmt.conversion(x)
+
+  s <- clone(x, copy.cache = FALSE, copy.storage = FALSE)
+  .set(s, "F", conversion(F))
+  class(s) <- .remove.ossa.pssa(class(s))
+
+  s
+}
+
+high.rank.rate <- function(F, rank, ssaobj = ssa(F), ...) {
+  s <- .clone.with.new.series(ssaobj, F)
+
+  # Decompose for computation leading singular values
+  .maybe.continue(s, groups = seq_len(rank), ...)
+
+  1 - sum(.sigma(s)[seq_len(rank)]^2) / wnorm(s)^2
 }
 
 pseudo.inverse <- function(A) {
@@ -74,13 +92,74 @@ orthogonalize <- function(Y, Z, sigma, side = c("bi", "left", "right"), normaliz
   }
 }
 
-.owcor <- function(X, L, LM, RM) {
-  fft.plan <- fft.plan.1d(nrow(X))
-  mx <- apply(X, 2,
-              function(v) {
-                h <- new.hmat(v, L = L, fft.plan = fft.plan)
-                as.vector(LM %*% hmatmul.wrap(h, t(RM), transposed = FALSE))
+# TODO Make emat really polymorphic class. Such wrappings are very ugly
+.traj.matrix <- function(x, ...)
+  UseMethod(".traj.matrix")
+
+.traj.matrix.1d.ssa <- .traj.matrix.toeplitz.ssa <- function(x, ...)
+  .get.or.create.hmat(x, ...)
+
+.traj.matrix.nd.ssa <- function(x, ...)
+  .get.or.create.hbhmat(x, ...)
+
+.traj.matrix.mssa <- function(x, ...)
+  .get.or.create.mhmat(x, ...)
+
+.traj.matrix.cssa <- function(x, ...)
+  .get.or.create.chmat(x, ...)
+
+.traj.matrix.ssa <- function(x, ...)
+  stop("`.traj.matrix' is not implemented for this kind of SSA")
+
+hbhmatmul.wrap <- function(h, mx, transposed = TRUE) {
+  if (ncol(mx) == 0)
+    return(matrix(NA_real_, if (transposed) hbhcols(h) else hbhrows(h), 0))
+
+  apply(mx, 2, hbhmatmul, hmat = h, transposed = transposed)
+}
+
+.traj.matrix.multiplier <- function(x, ...)
+  UseMethod(".traj.matrix.multiplier")
+
+.traj.matrix.multiplier.1d.ssa <- .traj.matrix.multiplier.toeplitz.ssa <- function(x, ...)
+  hmatmul.wrap
+
+.traj.matrix.multiplier.nd.ssa <- .traj.matrix.multiplier.mssa <- function(x, ...)
+  hbhmatmul.wrap
+
+.traj.matrix.multiplier.ssa <- function(x, ...)
+  stop("`.traj.matrix.multiplier' is not implemented for this kind of SSA")
+
+.frob.dist.series.lowrank <- function(F, sigma, Y, Z, ssaobj) {
+  # Get trajectory matrix multiplication function (facepalm)
+  matmul.wrap <- .traj.matrix.multiplier(ssaobj)
+
+  # This is efficient implementation of |T(F) - Y diag(sigma) Z^T|^2_F
+  s <- .clone.with.new.series(ssaobj, F)
+  TF <- .traj.matrix(s)
+
+  sZ <- Z * rep(sigma, each = nrow(Z))
+
+  res <- sum((sZ %*% crossprod(Y)) * sZ) + wnorm(s)^2 - 2*sum(matmul.wrap(TF, Y, transposed = TRUE) * sZ)
+
+  # Fix possible numeric error, force dist be nonnegative
+  if (res < 0)
+    res <- 0
+
+  res
+}
+
+.owcor <- function(Fs, LM, RM, ssaobj) {
+  # Get trajectory matrix multiplication function
+  matmul.wrap <- .traj.matrix.multiplier(ssaobj)
+
+  mx <- lapply(Fs,
+               function(F) {
+                 s <- .clone.with.new.series(ssaobj, F)
+                 TF <- .traj.matrix(s)
+                 as.vector(LM %*% matmul.wrap(TF, t(RM), transposed = FALSE))
               })
+  mx <- do.call(cbind, mx)
 
   # Compute covariations
   cov <- crossprod(mx)
@@ -107,7 +186,6 @@ print.iossa.result <- function(x, digits = max(3, getOption("digits") - 3), ...)
   cat("\nIterative O-SSA result:\n")
   cat("\tConverged:             ", ifelse(x$converged, "yes", "no"), "\n", sep = "")
   cat("\tIterations:            ", x$iter, "\n", sep = "")
-  cat("\tCondition numbers:     ", paste0(format(x$cond, digits = digits), collapse = ", "), "\n", sep = "")
   cat("\tInitial mean(tau):     ", format(mean(x$initial.tau), digits = digits), "\n", sep = "")
   cat("\tInitial tau:           ", paste0(format(x$initial.tau, digits = digits), collapse = ", "), "\n", sep = "")
   cat("\tI. O-SSA mean(tau):    ", format(mean(x$tau), digits = digits), "\n", sep = "")
@@ -153,24 +231,25 @@ summary.ossa <- function(object, digits = max(3, getOption("digits") - 3), ...)
   .set.decomposition(x, sigma = sigma, U = U, V = V)
 }
 
-.make.ossa.result <- function(x, Fs, ranks, IBL, IBR, initial.Fs, svd.method = "auto") {
-  L <- x$window
+.make.ossa.result <- function(x, Fs, ranks, idx, initial.Fs) {
+  IBL <- pseudo.inverse(.U(x)[, idx, drop = FALSE])
+  IBR <- pseudo.inverse(calc.v(x, idx))
 
   tau <- sapply(seq_along(Fs),
                 function(i) {
-                  high.rank.rate(Fs[[i]], L, ranks[i], svd.method = svd.method)
+                  high.rank.rate(Fs[[i]], ranks[i], ssaobj = x)
                 })
 
   initial.tau <- sapply(seq_along(initial.Fs),
                         function(i) {
-                          high.rank.rate(initial.Fs[[i]], L, ranks[i], svd.method = svd.method)
+                          high.rank.rate(initial.Fs[[i]], ranks[i], ssaobj = x)
                         })
 
   names(Fs) <- paste("F", seq_along(Fs), sep = "")
   out <- list(cond = c(Cond(IBL), Cond(IBR)),
-              owcor = .owcor(do.call(cbind, Fs), L, IBL, IBR),
-              wcor = wcor(do.call(cbind, Fs), L),
-              initial.wcor = wcor(do.call(cbind, initial.Fs), L),
+              owcor = .owcor(Fs, IBL, IBR, ssaobj = x), #MB just call owcor(x) here?
+              wcor = wcor(x, Fs = Fs),
+              initial.wcor = wcor(x, Fs = initial.Fs),
               tau = tau,
               initial.tau = initial.tau,
               initial.rec = initial.Fs)
@@ -179,13 +258,6 @@ summary.ossa <- function(object, digits = max(3, getOption("digits") - 3), ...)
   invisible(out)
 }
 
-.component.grouping <- function(elem.components, groups) {
-  lapply(groups,
-         function(group) {
-           rowSums(elem.components[, group, drop = FALSE])
-         })
-}
-
 svd2LRsvd <- function(d, u, v, basis.L, basis.R, need.project = TRUE, fast = TRUE) {
   rank <- length(d)
   # Check parameters' dims
@@ -250,9 +322,12 @@ iossa <- function(x, nested.groups, ..., tol = 1e-5, kappa = 2,
                   maxiter = 100,
                   norm = function(x) sqrt(mean(x^2)),
                   trace = FALSE,
-                  kappa.balance = 0.5,
-                  svd.method = "auto") {
-  N <- x$length; L <- x$window; K <- N - L + 1
+                  kappa.balance = 0.5) {
+  if (inherits(x, "cssa"))
+    stop("I-OSSA is not implemented for Complex SSA yet")
+
+  # Get mask
+  mask <- .hweights(x) > 0
 
   if (missing(nested.groups))
     nested.groups <- as.list(1:min(nsigma(x), nu(x)))
@@ -282,15 +357,14 @@ iossa <- function(x, nested.groups, ..., tol = 1e-5, kappa = 2,
     }
   }
 
-  # Grab the FFT plan
-  fft.plan <- .get.or.create.fft.plan(x)
-
   converged <- FALSE
   for (iter in seq_len(maxiter)) {
     lrss <- numeric(length(nested.groups))
     triples.list <- list()
     for (i in seq_along(nested.groups)) {
-      cur.ssa <- ssa(Fs[[i]], L, svd.method = svd.method, neig = min(ranks[i] + 1, L, K))
+      cur.ssa <- .clone.with.new.series(x, Fs[[i]])
+
+      .maybe.continue(cur.ssa, seq_len(ranks[i]))
       lrss[i] <- sum(cur.ssa$sigma[-seq_len(ranks[i])]^2)
       triples.list[[i]] <- .get.orth.triples(cur.ssa, seq_len(ranks[i]))
     }
@@ -321,22 +395,36 @@ iossa <- function(x, nested.groups, ..., tol = 1e-5, kappa = 2,
     dec <- svd2LRsvd(osigma, U, V, basU, basV, need.project = TRUE)
     sigma <- dec$sigma; Y <- dec$Y; Z <- dec$Z
 
-    elem.components <- .hankelize.multi.default(t(sigma * t(Y)), Z, fft.plan)
-    Fs.new <- .component.grouping(elem.components, nested.groups)
+    # Clone is used for hankelization
+    # We copy storage for saving quasi-hankel circulant, it is used for stored fft_plan
+    # TODO Separate circulants and plans just like in 1d-SSA
+    s <- clone(x, copy.cache = FALSE, copy.storage = TRUE)
+
+    # Add `ossa` class for prevent redecomposition
+    class(s) <- c("ossa", class(s)[class(s) != "ossa"])
+
+    .set.decomposition(s, sigma = sigma, U = Y, V = Z)
+    Fs.new <- reconstruct(s, groups = nested.groups)
 
     if (trace) {
-      Xs <- lapply(nested.groups,
-                   function(group) Y[, group, drop = FALSE] %*% (sigma[group] * t(Z[, group, drop = FALSE])))
-      svddist <- sapply(seq_along(Xs),
-                        function(i) sum((Xs[[i]] - hankel(Fs[[i]], L)) ^ 2))
-      cat(sprintf("SVDD(%d): %s\n", iter, paste0(sqrt(svddist), collapse = " ")))
+      svddist <- sapply(seq_along(nested.groups),
+                        function(i) .frob.dist.series.lowrank(Fs[[i]],
+                                                              sigma[nested.groups[[i]]],
+                                                              Y[, nested.groups[[i]], drop = FALSE],
+                                                              Z[, nested.groups[[i]], drop = FALSE],
+                                                              x))
 
-      nonhankeleness <- sapply(seq_along(Xs),
-                               function(i) sum((Xs[[i]] - hankel(Fs.new[[i]], L)) ^ 2))
-      cat(sprintf("NHNK(%d): %s\n", iter, paste0(sqrt(nonhankeleness), collapse = " ")))
+      nonhank <- sapply(seq_along(nested.groups),
+                        function(i) .frob.dist.series.lowrank(Fs.new[[i]],
+                                                              sigma[nested.groups[[i]]],
+                                                              Y[, nested.groups[[i]], drop = FALSE],
+                                                              Z[, nested.groups[[i]], drop = FALSE],
+                                                              x))
+      cat(sprintf("SVDD(%d): %s\n", iter, paste0(sqrt(svddist), collapse = " ")))
+      cat(sprintf("NHNK(%d): %s\n", iter, paste0(sqrt(nonhank), collapse = " ")))
     }
 
-    deltas <- sapply(seq_along(nested.groups), function(i) norm(Fs.new[[i]] - Fs[[i]]))
+    deltas <- sapply(seq_along(nested.groups), function(i) .series.dist(Fs.new[[i]], Fs[[i]], norm, mask))
     if (max(deltas) < tol) {
       converged <- TRUE
       Fs <- Fs.new
@@ -346,9 +434,6 @@ iossa <- function(x, nested.groups, ..., tol = 1e-5, kappa = 2,
     Fs <- Fs.new
   }
 
-  IBL <- pseudo.inverse(Y)
-  IBR <- pseudo.inverse(Z)
-
   x <- clone(x, copy.cache = FALSE) # TODO Maybe preserve relevant part of cache?
   .save.oblique.decomposition(x, sigma, Y, Z, idx)
 
@@ -360,7 +445,7 @@ iossa <- function(x, nested.groups, ..., tol = 1e-5, kappa = 2,
   # Save call info
   x$call <- match.call()
 
-  out <- .make.ossa.result(x, Fs, ranks, IBL, IBR, rec, svd.method = svd.method)
+  out <- .make.ossa.result(x, Fs, ranks, idx, rec)
   out[c("iter", "converged", "kappa", "maxiter", "tol", "call")] <-
       list(iter, converged, kappa, maxiter, tol, match.call())
 
@@ -394,7 +479,7 @@ iossa <- function(x, nested.groups, ..., tol = 1e-5, kappa = 2,
   invisible(x)
 }
 
-fossa <- function(x, nested.groups, FILTER = diff, gamma = 1, ...) {
+fossa <- function(x, nested.groups, FILTER = diff, gamma = 1, normalize = FALSE, ...) {
   if (!is.function(FILTER)) {
     FILTER.coeffs <- FILTER
     FILTER <- function(x) {
@@ -417,8 +502,10 @@ fossa <- function(x, nested.groups, FILTER = diff, gamma = 1, ...) {
 
   Z <- V * rep(osigma, each = nrow(V))
 
-  fZ <- apply(Z, 2, FILTER)
-  dec <- eigen(crossprod(rbind(Z, gamma * fZ)), symmetric = TRUE)
+  Y <- if (normalize) V else Z
+
+  fY <- apply(Y, 2, FILTER)
+  dec <- eigen(crossprod(rbind(Y, gamma * fY)), symmetric = TRUE)
   U <- U %*% dec$vectors
   Z <- Z %*% dec$vectors
   sigma <- rep(1, ncol(U))
@@ -499,14 +586,14 @@ owcor <- function(x, groups, ..., cache = TRUE) {
                  paste0(basis, collapse = ", ")))
 
   # Compute reconstruction.
-  F <- reconstruct(x, groups, ..., cache = cache)
+  Fs <- reconstruct(x, groups, ..., cache = cache)
 
   LM <- pseudo.inverse(.U(x)[, basis, drop = FALSE]) # No colspan here!!!!
   RM <- pseudo.inverse(calc.v(x, basis)) # No rowspan here!!!!
 
   # Compute oblique w-correlations and return
-  res <- .owcor(do.call(cbind, F), L = x$window, LM, RM)
-  colnames(res) <- rownames(res) <- names(F)
+  res <- .owcor(Fs, LM, RM, ssaobj = x)
+  colnames(res) <- rownames(res) <- names(Fs)
 
   res
 }
